module fm_top(clk_pad, pb, ss, led, sled, rxd, banksel,
	ramadr, ramdata, ramce_n, ramoe_n, ramwe_n, rambe_n,
	tx, txd, din, rclk, init_b, cs, sclk, sdo, sdi);
	input clk_pad, din, sdi, rxd;
	input [3:0] pb, banksel;
	input [7:0] ss;
	output [7:0] led;
	output [11:0] sled;
	output [17:0] ramadr;
	inout [31:0] ramdata;
	output [1:0] ramce_n;
	output ramoe_n, ramwe_n;
	output [3:0] rambe_n;
	output tx, txd, rclk, init_b, cs, sclk, sdo;
	wire clk, clk1, clk_audio, locked1, locked2;
	dcm1 dcm1(.CLKIN_IN(clk_pad), .CLK0_OUT(clk), .CLKFX_OUT(clk1), .LOCKED_OUT(locked1));
	dcm2 dcm2(.CLKIN_IN(clk1), .RST_IN(~locked1), .CLKFX_OUT(clk_audio), .LOCKED_OUT(locked2));
	wire reset = ~locked1;
	wire [15:0] cpu_adr;
	wire [14:0] ioadr;
	wire cpu_rd, cpu_memwr, cpu_iowr, iec;
	wire [3:0] cpu_be;
	wire [31:0] cpu_data_in, cpu_data_out;
	reg int = 0;
	kx_mips kx_mips(.clk(clk), .reset(reset), .int(int),
		.adr(cpu_adr), .ioadr(ioadr), 
		.rd(cpu_rd), .memwr(cpu_memwr), .iowr(cpu_iowr), .be(cpu_be), 
		.data_in(cpu_data_in), .data_out(cpu_data_out),
		.valid(), .txd(txd), .iec(iec));
	assign ramce_n = 2'b00;
	assign ramadr = { 2'b0, cpu_adr };
	assign ramdata = cpu_rd ? 32'hZZZZZZZZ : cpu_data_out;
	assign ramoe_n = ~cpu_rd;
	assign ramwe_n = ~cpu_memwr;
	assign rambe_n = ~cpu_be;
	reg [6:0] port;
	reg [11:0] sled = 'hfff;
	reg sdo, sclk, cs, rclk, init_b;
	always @(posedge clk) 
		if (reset) port <= 0;
		else if (cpu_iowr) begin
			if (~&ioadr[14:13]) port <= cpu_data_out[6:0];
			if (ioadr[14]) 
				if (ioadr[13]) sled <= cpu_data_out[11:0];
				else { init_b, rclk, cs, sclk, sdo } <= cpu_data_out[4:0];
		end
	assign led = { tx, port };
	reg [15:0] timer;
	wire [16:0] timer_d = timer + 1;
	always @(posedge clk) timer <= timer_d[15:0];
	always @(posedge clk) begin
		if (~iec) int <= 0;
		else if (timer_d[16]) int <= 1;
	end
	wire busy;
	assign cpu_data_in = cpu_adr[15] ? cpu_adr[14] ? { 25'b0, pb, rxd, din, sdi } : { timer, busy, 3'b0, banksel, ss } : ramdata;
	wire next, lr, o_write;
	wire [31:0] fm_out, buf_out;
	wire [4:0] n;
	fm fm(.clk(clk), .adr(ioadr[13:0]), .data(cpu_data_out), 
		.write(cpu_iowr & ~ioadr[14]), .busy(busy), 
		.o(fm_out), .o_write(o_write), .fifo_full(&n));
	fifo fifo(.ainit(reset), 
		.din(fm_out), .wr_en(o_write), .wr_clk(clk), .wr_count(n),
		.dout(buf_out), .rd_en(next), .rd_clk(clk_audio));
	dait dait(.clk(clk_audio), .reset(~locked2), .next(next), .lr(lr), 
		.data(lr ? buf_out[31:16] : buf_out[15:0]), .tx(tx));
endmodule

module kx_mips(clk, reset, int, adr, ioadr, rd, memwr, iowr, be, data_in, data_out, valid, iec, txd);
	input clk, reset, int;
	input [31:0] data_in;
	output rd, memwr, iowr, valid, iec, txd;
	output [3:0] be;
	output [15:0] adr;
	output [14:0] ioadr;
	output [31:0] data_out;
endmodule

module fm(clk, adr, data, write, busy, o, o_write, fifo_full);
	input clk, write, fifo_full;
	input [13:0] adr;
	input [31:0] data;
	output busy, o_write;
	output [31:0] o;
endmodule

module dait(clk, reset, data, next, lr, tx);
	input clk, reset;
	input [15:0] data;
	output next, lr, tx;
endmodule

