	.text
	.align	4
	.set	noreorder
	.set	noat
	.globl	_start

	SIM	=	0
	CARD	=	1
	BOOTABLE	=	1
	FLASH_BOOT	=	0
	SERIAL_BOOT	=	1
	NATIVE_DIV	=	0
	CLOCK_FREQ	=	50
	BPS	=	115200
	BUFN	=	0x1000

# global variables

	G_FREECH	=	0
	G_FREECHTAIL	=	4
	G_ACTIVECH	=	8

	G_FUNC	=	0x10
	G_MIDIWORKPTR	=	0x14
	G_MIDICH	=	0x18
	G_STATUS	=	0x1a
	G_VV	=	0x1b
	G_META	=	0x1c

	G_TEMPO	=	0x20
	G_TIMEBASE	=	0x24	# 2byte

	G_TARGETTIME	=	0x28
	G_CURTIME	=	0x2c	# 2byte

	G_MIDIPTR	=	0x30
	G_MIDILIM	=	0x34
	G_TONE	=	0x38

	G_FILENUM	=	0x3c

	G_FAT	=	0x40
	G_RDE	=	0x44
	G_FILEOFS	=	0x48
	G_FILELEN	=	0x4c
	G_CARDADR	=	0x50
	G_CLUSTERMASK	=	0x54
	G_CLUSTER	=	0x56

	G_BUFWP	=	0x58
	G_BUFRP	=	0x5c
	G_BUFN	=	0x60

	G_KEY	=	0x64
	G_KEYON	=	0x65
	G_STARTED	=	0x66

	G_DIGIT	=	0x7b
	G_7SEG	=	0x7c

	G_SIZE	=	0x80

# structure

	OP_MUL	=	0
	OP_DET	=	2
	OP_PG1U	=	3
	OP_SIZE	=	4
	OP_N	=	8

	CH_FM	=	4
	CH_BASE	=	8
	CH_BEND	=	10
	CH_ID	=	12
	CH_PR	=	14
	CH_VELO	=	16
	CH_PROG	=	17
	CH_NOTE	=	18
	CH_FLAG	=	19
	CH_PAN	=	20
	CH_VOLUME	=	21
	CH_ON	=	22
	CH_PERC	=	23
	CH_OPOFS	=	24
	CH_SIZE	=	CH_OPOFS+OP_N*OP_SIZE
	CH_N	=	32

# tone data format

	TONE_FLAGS	=	0
	TONE_CON	=	1
	TONE_ON	=	2
	TONE_LFO	=	4
	TONE_SN	=	8
	TONE_SC	=	9
	TONE_HEADER_SIZE	=	16

	PARAM_MUL	=	0
	PARAM_ML	=	2
	PARAM_CON	=	4
	PARAM_DET	=	5
	PARAM_VS	=	6
	PARAM_PH	=	7
	PARAM_AT1	=	8
	PARAM_AT2	=	9
	PARAM_DT1	=	10
	PARAM_DT2	=	11
	PARAM_S	=	12
	PARAM_RT1	=	13
	PARAM_RT2	=	14
	PARAM_HT	=	15
	PARAM_AL	=	16
	PARAM_TL	=	17
	PARAM_DL	=	18
	PARAM_SL	=	19
	PARAM_RL	=	20
	PARAM_SIZE	=	22

	MASK_PERC	=	8
	MASK_SS	=	0x10
	MASK_NZ	=	0x20
	MASK_CON	=	0x80

	MIDI_PROG	=	0
	MIDI_NOTE	=	1
	MIDI_VELO	=	2
	MIDI_VOLUME	=	3
	MIDI_EXPRESSION	=	4
	MIDI_PAN	=	5
	MIDI_BEND	=	6
	MIDI_BENDSEN	=	7
	MIDI_RPN	=	8
	MIDI_SIZE	=	12
	MIDI_N	=	16

# work address [Attention] absolute address defined

	CHWORK	=	0x2700
	CHWORK_SIZE	=	CH_SIZE*CH_N
	MIDIWORK	=	CHWORK+CHWORK_SIZE
	MIDIWORK_SIZE	=	MIDI_SIZE*MIDI_N
	GVAR	=	MIDIWORK+MIDIWORK_SIZE	# G_SIZE byte
	TONEADR	=	0x3000
	BUF	=	0x10000	# BUFN bytes

# I/O address

	IOBASE	=	0x20000
	TIMER	=	IOBASE+2
	FM_BASE	=	IOBASE
	IOCARD	=	IOBASE+0x10000
	IO_PBUTTON	=	IOBASE+0x10000
	IO_FLASH	=	IOBASE+0x10000
	IO_SERIAL	=	IOBASE+0x10000
	IO_7SEG	=	IOBASE+0x18000

	FM_RATE_A1	=	0x1c
	FM_RATE_A2	=	0x18
	FM_RATE_D1	=	0x14
	FM_RATE_D2	=	0x10
	FM_RATE_S	=	0xc
	FM_RATE_R1	=	8
	FM_RATE_R2	=	4
	FM_RATE_H	=	0
	FM_L_A1	=	0x3c
	FM_L_A2	=	0x38
	FM_L_D1	=	0x34
	FM_L_D2	=	0x30
	FM_L_S	=	0x2c
	FM_L_R1	=	0x28
	FM_L_R2	=	0x24
	FM_L_H	=	0x20
	FM_PHASE	=	0x40
	FM_PG0	=	0x60
	FM_PG1	=	0x64
	FM_COUNT	=	0x68
	FM_LFO	=	0x70
	FM_VOL	=	0x74
	FM_SIZE_LOG	=	7	# 0x80bytes
	FM_SIZE	=	1<<FM_SIZE_LOG
	FM_CHOFS	=	OP_N*FM_SIZE

	PB_RESET	=	8
	PB_PREV	=	4
	PB_REPLAY	=	2
	PB_NEXT	=	1
	PB_ALL	=	PB_RESET|PB_REPLAY|PB_PREV|PB_NEXT

	.macro	mov	dst,src
	addu	\dst,zero,\src
	.endm

	FLASH_OE_POS	=	4
	FLASH_RCLK_POS	=	3
	FLASH_DIN_POS	=	1
	SERIAL_POS	=	2

_start:
	li	a0,GVAR
	li	v0,G_SIZE/4-1
0:
	sw	zero,0(a0)
	addiu	a0,a0,4
	bne	v0,zero,0b
	addiu	v0,v0,-1
	li	gp,GVAR
	li	v0,1
	j	start
	mtc0	v0,$12	# interrupt enable

# interrupt handler
# using k0,k1,fp

	.org	0x80

	lbu	k0,G_DIGIT(gp)
	nop
	addiu	k0,k0,1
	andi	k0,k0,3
	sb	k0,G_DIGIT(gp)
	li	k1,0x100
	sllv	k1,k1,k0
	addu	fp,gp,k0
	lbu	k0,G_7SEG(fp)
	nop
	or	k1,k1,k0
	nor	k1,k1,k1
	li	k0,IO_7SEG
	sw	k1,0(k0)
	eret

start:
	.if	!SIM
	.if	FLASH_BOOT
	li	a1,0x71386d76
	sw	a1,G_7SEG(gp)
	li	a0,IO_FLASH
	li	t0,1<<FLASH_OE_POS|1<<FLASH_RCLK_POS
	li	t1,1<<FLASH_OE_POS
	sb	t1,0(a0)
	li	v1,0
	li	t2,0x6e795355
1:
	sb	t0,0(a0)
	lbu	v0,0(a0)
	srl	v1,v1,1
	andi	v0,v0,1<<FLASH_DIN_POS
	sll	v0,v0,31-FLASH_DIN_POS
	sb	t1,0(a0)
	bne	v1,t2,1b
	or	v1,v1,v0
	li	a1,0	#adr
	li	a2,0x10000-4	#lim
2:
	li	a3,32	#count
3:
	sb	t0,0(a0)
	lbu	v0,0(a0)
	srl	v1,v1,1
	andi	v0,v0,1<<FLASH_DIN_POS
	sll	v0,v0,31-FLASH_DIN_POS
	sb	t1,0(a0)
	addiu	a3,a3,-1
	bne	a3,zero,3b
	or	v1,v1,v0
	sw	v1,0(a1)
	bne	a1,a2,2b
	addiu	a1,a1,4
	.endif

	.if	SERIAL_BOOT
	li	t3,0
	li	t4,0x6d795030
	sw	t4,G_7SEG(gp)
loadhex:
	jal	recv
	nop
	li	at,':'
	bne	v0,at,loadhex
	nop
	addiu	t3,t3,1
	andi	at,t3,0xf
	beql	at,zero,2f
	xori	t4,t4,0x80
2:
	jal	recvd
	sw	t4,G_7SEG(gp)
	jal	recvd
	sll	t0,v0,4
	jal	recvd
	or	t0,t0,v0	# count
	jal	recvd
	sll	t1,v0,12
	sll	v0,v0,8
	jal	recvd
	or	t1,t1,v0
	sll	v0,v0,4
	jal	recvd
	or	t1,t1,v0
	jal	recvd
	or	t1,t1,v0	# adr
	jal	recvd
	nop
	li	at,1
	beq	v0,at,run
	nop
	bne	v0,zero,loadhex
	nop
1:
	jal	recvd
	nop
	jal	recvd
	sll	t2,v0,4
	or	t2,t2,v0	# data
	sb	t2,0(t1)
	addiu	t0,t0,-1
	bne	t0,zero,1b
	addiu	t1,t1,1
	j	loadhex
	nop

# out: v0 (0-9)
# use: at,v0,v1,a0-a3

recvd:
	mov	v1,ra
	jal	recv
	nop
	addiu	at,v0,-'0'
	bltz	at,2f
	addiu	at,v0,-'9'
	bgtzl	at,1f
	addiu	at,v0,-'A'
	jr	v1
	addiu	v0,v0,-'0'
1:
	bltz	at,2f
	addiu	at,v0,-'F'
	bgtz	at,2f
	nop
	jr	v1
	addiu	v0,v0,10-'A'
2:
	jr	v1
	li	v0,0

# use: v0,a0-a3
# out: v0 data
# 4*timing+24 clocks

recv:
	mov	a3,ra
	li	a1,IO_SERIAL
1:
	lbu	a0,0(a1)
	nop
	andi	a0,a0,1<<SERIAL_POS
	bne	a0,zero,1b
	li	a2,7-1
	li	a0,3*(1000000*CLOCK_FREQ/BPS-12)/4-1
	jal	sample1
	li	v0,0
2:
	jal	sample
	nop
	bne	a2,zero,2b
	addiu	a2,a2,-1
	jal	sample
	mov	a2,v0
	jr	a3
	mov	v0,a2

# in:  a0 timing
#      a1 address
#      v0 last data
# out: v0 data
# use: v0,a0
# 2*timing+8 clocks

sample:
	li	a0,(1000000*CLOCK_FREQ/BPS-12)/2-1
sample1:
	bne	a0,zero,sample1
	addiu	a0,a0,-1
	lbu	a0,0(a1)
	srl	v0,v0,1
	andi	a0,a0,1<<SERIAL_POS
	sll	a0,a0,7-SERIAL_POS
	jr	ra
	or	v0,v0,a0

	.else
	j	run
	nop
	.endif
	.else
	j	run
	nop
	.endif

	.org	0x800

run:
	li	a0,IOBASE
	lhu	a0,0(a0)
	li	gp,GVAR
	andi	a0,a0,0xfff
	xori	a0,a0,0xf00	# upper
	sh	a0,G_FILENUM(gp)
restart:
	jal	Init
	nop
	jal	FMStop
	nop
	jal	NoteOffAll
	nop
	.if	SIM
	la	a0,tone+4
	sw	a0,G_TONE(gp)
	.else
	jal	FMStart
	nop
	li	a0,TONEADR
	li	v1,0x454e4f54
	lw	v0,0(a0)
1:
	addiu	a0,a0,4
	bnel	v0,v1,1b
	lw	v0,0(a0)
	sw	a0,G_TONE(gp)
	.endif
	.if	SIM
	li	a0,CHWORK
	sw	a0,G_ACTIVECH(gp)
	sw	zero,0(a0)
	sh	zero,CH_ID(a0)
	li	a1,0x1c
	li	a2,69
	li	a3,0x7f
	jal	ChannelOn
	nop
	jal	FMStart
	nop
2:
	j	2b
	nop
	.endif
	jal	MidiInit
	nop
	jal	FileInit
	nop
	lhu	a0,G_FILENUM(gp)
	jal	FileOpen
	nop
	jal	FillBuffer
	nop
	jal	SMFHeader
	nop
	j	SMFPlay
	nop

Init:
	li	sp,IOBASE
	li	gp,GVAR
	li	t0,CHWORK
	sw	t0,G_FREECH(gp)
	li	t1,CH_N
	li	t2,FM_BASE
1:
	addiu	t3,t0,CH_SIZE
	sw	t3,0(t0)
	sw	t2,CH_FM(t0)
	mov	t0,t3
	addiu	t1,t1,-1
	bne	t1,zero,1b
	addiu	t2,t2,FM_CHOFS
	addiu	t0,t0,-CH_SIZE
	sw	zero,0(t0)
	sw	t0,G_FREECHTAIL(gp)
	sw	zero,G_ACTIVECH(gp)
	jr	ra
	sb	zero,G_STARTED(gp)

# use: a0

FMStart:
	li	a0,FM_BASE+0x8000
	li	at,1
	jr	ra
	sb	at,0(a0)

# use: a0,a1

FMStop:
	li	a0,FM_BASE+0x8000
	sb	zero,0(a0)
	li	a0,FM_BASE
	lhu	a1,0(a0)
1:
	nop
	andi	a1,a1,0x8000
	bnel	a1,zero,1b
	lhu	a1,0(a0)
	jr	ra
	nop

# void KeyOn(u8 prog, u8 note, u8 velo, u8 volume, u16 id, u8 pan, s16 bend);
# in:  a0 prog
#      a1 note
#      a2 velo
#      a3 volume
#      v0 id
#      v1 pan
#      t7 bend

KeyOn:
	lw	t0,G_ACTIVECH(gp)
	j	6f
	addiu	t1,gp,G_ACTIVECH
	lhu	t2,CH_ID(t0)
7:
	nop
	beql	t2,v0,8f
	lw	t2,0(t0)
	mov	t1,t0
	lw	t0,0(t0)
	nop
6:
	bnel	t0,zero,7b
	lhu	t2,CH_ID(t0)
	lw	t0,G_FREECH(gp)
	nop
	beql	t0,zero,1f
	lw	t0,G_ACTIVECH(gp)	# p
	lw	t1,0(t0)
	nop
	bne	t1,zero,2f
	sw	t1,G_FREECH(gp)
	j	2f
	sw	zero,G_FREECHTAIL(gp)
1:
	nop
	beq	t0,zero,9f
	nop
	lw	t1,0(t0)
	j	2f
	sw	t1,G_ACTIVECH(gp)
8:
	nop
	sw	t2,0(t1)
2:
	multu	a2,a3
	mflo	t4
	lw	t2,G_ACTIVECH(gp)	# pn
	j	3f
	addiu	t1,gp,G_ACTIVECH	# p0
9:
	jr	ra
	nop
4:
	lw	t2,0(t2)
	nop
3:
	beql	t2,zero,5f
	sw	t0,0(t1)
	lhu	t3,CH_PR(t2)
	nop
	subu	at,t4,t3
	bgezl	at,4b
	mov	t1,t2
	sw	t0,0(t1)
5:
	sw	t2,0(t0)
	sh	t4,CH_PR(t0)
	sh	t7,CH_BEND(t0)
	sb	v1,CH_PAN(t0)
	sh	v0,CH_ID(t0)
	sb	a3,CH_VOLUME(t0)
	sb	a2,CH_VELO(t0)
	sb	a1,CH_NOTE(t0)
	sb	a0,CH_PROG(t0)
	li	a0,1
	sb	a0,CH_FLAG(t0)
	j	ChannelForce
	mov	a0,t0

# void ChannelOn(Channel *ch, u8 prog, u8 note, u8 velo);
# in: a0 ch
#     a1 prog
#     a2 note
#     a3 velo
# use: at,v0,v1,a0-a3,t0-t9

	.macro	egid	dst,src
	lbu	t2,\src(t9)
	lbu	t3,\src+PARAM_SIZE(t9)
	multu	t2,t0
	mflo	t2
	multu	t3,t1
	mflo	t3
	addu	t2,t2,t3
	sh	t2,\dst(s4)	#bit15:7
	.endm

	.macro	egidl	src
	lbu	t2,\src(t9)
	lbu	t3,\src+PARAM_SIZE(t9)
	multu	t2,t0
	mflo	t2
	multu	t3,t1
	mflo	t3
	addu	t2,t2,t3
	.endm

	.macro	pgid8	src
	lbu	t2,\src(t9)
	lbu	t3,\src+PARAM_SIZE(t9)
	multu	t0,t2
	mflo	t2
	multu	t1,t3
	mflo	t3
	addu	t2,t2,t3	#bit15:8
	.endm

	.macro	pgid16	src
	lhu	t2,\src(t9)
	lhu	t3,\src+PARAM_SIZE(t9)
	multu	t0,t2
	mflo	t2
	multu	t1,t3
	mflo	t3
	addu	t2,t2,t3	#bit23:8
	.endm

ChannelOn:
	sw	s0,-4(sp)
	sw	s1,-8(sp)
	sw	s2,-12(sp)
	sw	s3,-16(sp)
	sw	s4,-20(sp)
	addiu	sp,sp,-20
	lw	v0,G_TONE(gp)
	mov	s0,a0	# &ch
	lhu	v1,0(v0)
	sll	a1,a1,1
	addu	v0,v0,v1
	addu	a1,v0,a1
	lhu	a1,0(a1)
	nop
	addu	t9,v0,a1	# &tone
	lw	s4,CH_FM(s0)
	mov	t6,ra
	lw	t0,TONE_LFO(t9)
	lbu	t8,TONE_SN(t9)
	sw	t0,FM_LFO(s4)
	.if SIM
	li	t0,0x8080
	sh	t0,FM_VOL(s4)
	.else
	jal	ChannelVolume
	nop
	.endif
	li	t0,PARAM_SIZE
	multu	t8,t0
	lbu	t7,TONE_ON(t9)
	lbu	s3,TONE_FLAGS(t9)
	lbu	s2,TONE_CON(t9)
	andi	at,s3,MASK_PERC
	sb	at,CH_PERC(s0)
	sb	t7,CH_ON(s0)
	li	t1,0
	addiu	t5,t9,TONE_SC	# &sc[i]
	addiu	t9,t9,TONE_HEADER_SIZE
	mflo	t8	# PARAM_SIZE * sN
	beq	t8,zero,1f
	li	t2,0	# PARAM_SIZE * i
	lbu	t3,0(t5)
3:
	addiu	t5,t5,1
	subu	at,a2,t3
	bltz	at,2f
	nop
	addiu	t2,t2,PARAM_SIZE
	subu	at,t2,t8
	bltzl	at,3b
	lbu	t3,0(t5)
	addu	t9,t9,t8
	b	1f
	addiu	t9,t9,-PARAM_SIZE
2:
	beql	t2,zero,8f
	li	t0,127
	addu	t9,t9,t2
	addiu	t9,t9,-PARAM_SIZE
	lbu	t0,-2(t5)
	lbu	t5,-1(t5)
	subu	t2,a2,t0
	.if NATIVE_DIV
	sll	a0,t2,8
	subu	a1,t5,t0
	.dc.l	4<<21|5<<16|0x1b	# divu
	mflo	t1
	.else
	sll	a0,t2,8
	jal	Div
	subu	a1,t5,t0
	mov	t1,v0	# ratio
	.endif
1:
	li	t0,127
8:
	subu	a3,t0,a3
	andi	at,s3,MASK_SS
	beql	at,zero,5f
	sll	s1,a2,1
	addiu	s1,a2,60
5:
	beql	t8,zero,6f
	li	t8,PARAM_SIZE
6:
	sll	s1,s1,7
	sh	s1,CH_BASE(s0)
	li	t0,0x100
	subu	t0,t0,t1
	lh	t5,CH_BEND(s0)
	addiu	s0,s0,CH_OPOFS
7:
	sll	v1,s2,7
	or	v1,v1,s3
# s4 op (I/O address)
# t9 p
# t0&t1 ratio
# s1 note << 7
# a3 velo
# v1 bit7:carrier bit6:fz bit5:nz
# EG part
	egid	FM_RATE_A1,PARAM_AT1
	egid	FM_RATE_A2,PARAM_AT2
	egid	FM_RATE_D1,PARAM_DT1
	egid	FM_RATE_D2,PARAM_DT2
	egid	FM_RATE_S,PARAM_S
	egid	FM_RATE_R1,PARAM_RT1
	egid	FM_RATE_R2,PARAM_RT2
	egid	FM_RATE_H,PARAM_HT
	egidl	PARAM_AL
	ori	t2,t2,1
	sh	t2,FM_L_A1(s4)
	addu	t4,zero,t2
	egidl	PARAM_TL
	andi	t2,t2,0xfffe
	sltu	t3,t4,t2
	or	t2,t2,t3
	sh	t2,FM_L_A2(s4)
	addu	t4,zero,t2
	egidl	PARAM_DL
	andi	t2,t2,0xfffe
	sltu	t3,t4,t2
	or	t2,t2,t3
	sh	t2,FM_L_D1(s4)
	addu	t4,zero,t2
	egidl	PARAM_SL
	andi	t2,t2,0xfffe
	sltu	t3,t4,t2
	or	t2,t2,t3
	sh	t2,FM_L_D2(s4)
	sh	zero,FM_L_S(s4)
	egidl	PARAM_RL
	andi	t2,t2,0xfffe
	sh	t2,FM_L_R1(s4)
	sh	zero,FM_L_R2(s4)
	li	t3,0xffff
	sh	t3,FM_L_H(s4)
	li	t3,8
	sh	t3,FM_PHASE(s4)
# PG part
	andi	at,v1,MASK_CON
	beq	at,zero,11f
	mov	t4,a3
	pgid8	PARAM_VS
	multu	t2,t4
	mflo	t2
	srl	t4,t2,7+8
11:
	pgid16	PARAM_ML
	or	t2,t2,t4
	lbu	t3,PARAM_CON(t9)
	sll	t2,t2,8
	or	t2,t2,t3
	sw	t2,FM_PG0(s4)
	pgid8	PARAM_DET
	srl	t2,t2,8
	sb	t2,OP_DET(s0)
	pgid16	PARAM_MUL
	srl	t2,t2,8
	sh	t2,OP_MUL(s0)
	sb	v1,OP_PG1U(s0)
	mov	a0,s0
	jal	CalcPG1
	addu	a1,s1,t5
	sw	v0,FM_PG1(s4)
	pgid8	PARAM_PH
	srl	t2,t2,8
	sll	t2,t2,14
	sw	t2,FM_COUNT(s4)
#
	addiu	t7,t7,-1
	addiu	s0,s0,OP_SIZE
	srl	s2,s2,1
	andi	s3,s3,~MASK_NZ&0xff	# noise is first operator only
	addiu	s4,s4,FM_SIZE
	bne	t7,zero,7b
	addu	t9,t9,t8
	lw	s4,0(sp)
	lw	s3,4(sp)
	lw	s2,8(sp)
	lw	s1,12(sp)
	lw	s0,16(sp)
	jr	t6
	addiu	sp,sp,20

# in:  a0 Operator *
#      a1 (note << 8) + bend
# out: v0 PG1 value
# use: at,a1,a2

CalcPG1:
	lb	a2,OP_DET(a0)
	nop
	addu	a1,a1,a2	# v
	srl	a2,a1,7	# index
	addiu	at,a2,-510
	bgezl	at,1f
	li	a2,510
1:
	sll	a2,a2,2
	lw	v0,deltatable+4(a2)
	andi	a1,a1,0x7f
	multu	v0,a1
	li	v0,128
	subu	a1,v0,a1
	lw	v0,deltatable(a2)
	mflo	a2
	multu	v0,a1
	mflo	v0
	addu	a2,a2,v0
	lhu	a1,OP_MUL(a0)
	srl	a2,a2,7	# d
	multu	a2,a1
	mflo	a2
	lbu	a1,OP_PG1U(a0)
	srl	a2,a2,9	# delta
	sll	a1,a1,24
	jr	ra
	or	v0,a2,a1

# in:  a0 &channel
# use: at,v0,v1,a1

ChannelVolume:
	lbu	v1,CH_VOLUME(a0)
	lbu	v0,CH_PAN(a0)
	li	a1,0x7f
	subu	a1,a1,v0
	addiu	at,v0,-0x40
	bgezl	at,8f
	li	v0,0x40
8:
	multu	v0,v1
	mflo	v0
	sll	v0,v0,2
	andi	v0,v0,0xff00
	addiu	at,a1,-0x40
	bgezl	at,9f
	li	a1,0x40
9:
	multu	a1,v1
	mflo	a1
	srl	a1,a1,6
	andi	a1,a1,0xff
	or	v0,v0,a1
	lw	a1,CH_FM(a0)
	jr	ra
	sh	v0,FM_VOL(a1)

# void KeyOff(u16 id);
# in:  a0 id
# use: v0,v1,a0,t0-t3

KeyOff:
	lw	t1,G_ACTIVECH(gp)	# p
	j	1f
	addiu	t0,gp,G_ACTIVECH	# p0
9:
	jr	ra
	nop
2:
	lw	t1,0(t1)
	nop
1:
	beq	t1,zero,9b
	nop
	lhu	t2,CH_ID(t1)
	nop
	bnel	t2,a0,2b
	mov	t0,t1
	lbu	t2,CH_PERC(t1)
	nop
	bne	t2,zero,9b
	nop
	lw	t2,0(t1)
	lw	t3,G_FREECHTAIL(gp)
	sw	t2,0(t0)
	beql	t3,zero,4f
	sw	t1,G_FREECH(gp)
	sw	t1,0(t3)
4:
	sw	t1,G_FREECHTAIL(gp)
	sw	zero,0(t1)
	mov	a0,t1
# fall to ChannelOff

# void ChannelOff(Channel *ch);
# in: a0 ch
# use: v0,v1,a0

ChannelOff:
	lw	a0,CH_FM(a0)
	li	v0,OP_N
	li	v1,2
1:
	sb	v1,FM_PHASE(a0)
	addiu	v0,v0,-1
	bne	v0,zero,1b
	addiu	a0,a0,FM_SIZE
	jr	ra
	nop

# in: a0 ch
# use: v0,v1,a0,a1

ChannelForce:
	lw	a0,CH_FM(a0)
	li	v0,OP_N
	li	v1,1
	li	a1,0xff00
1:
	sb	v1,FM_PHASE(a0)
	sh	a1,FM_RATE_R2(a0)
	addiu	v0,v0,-1
	bne	v0,zero,1b
	addiu	a0,a0,FM_SIZE
	jr	ra
	nop

NoteOffAll:
	li	a0,FM_BASE
	li	v0,OP_N*CH_N
	li	v1,2
1:
	sb	v1,FM_PHASE(a0)
	addiu	v0,v0,-1
	bne	v0,zero,1b
	addiu	a0,a0,FM_SIZE
	jr	ra
	nop

# in:  a0 id_low
#      a1 bend
# use: at,v0,a1,a2,t0-t7

Bend:
	mov	t7,ra
	lw	t0,G_ACTIVECH(gp)
	sll	t3,a1,1	# bend
	j	1f
	mov	t2,a0	# id_low
	lbu	t1,CH_ID(t0)
2:
	nop
	bne	t1,t2,4f
	addiu	a0,t0,CH_OPOFS
	lhu	a1,CH_BASE(t0)
	lw	t6,CH_FM(t0)
	addu	t4,a1,t3
	lbu	t5,CH_ON(t0)
3:
	jal	CalcPG1
	mov	a1,t4
	sw	v0,FM_PG1(t6)
	addiu	a0,a0,OP_SIZE
	addiu	t5,t5,-1
	bne	t5,zero,3b
	addiu	t6,t6,FM_SIZE
4:
	lw	t0,0(t0)
	nop
1:
	bnel	t0,zero,2b
	lbu	t1,CH_ID(t0)
	jr	t7
	nop

# in:  a0 id_low
#      a1 volume*expression>>7
# use: at,v0,v1,a0,a1,t0-t4

ChangeVolume:
	mov	t4,ra
	lw	t0,G_ACTIVECH(gp)
	mov	t3,a1	# volume*expresstion>>7
	j	1f
	mov	t2,a0	# id_low
2:
	nop
	bnel	t1,t2,4f
	lw	t0,0(t0)
	sb	t3,CH_VOLUME(t0)
	jal	ChannelVolume
	mov	a0,t0
	lw	t0,0(t0)
4:
	nop
1:
	bnel	t0,zero,2b
	lbu	t1,CH_ID(t0)
	jr	t4
	nop

######## MIDI ########

# use: v0,v1,a0,a1

MidiInit:
	li	a0,MIDIWORK
	li	a1,MIDI_N
	lui	v0,0x6400
	lui	v1,0x200
	ori	v1,v1,0x407f
	li	a2,0x7f
1:
	sw	v0,0(a0)
	sw	v1,4(a0)
	sw	a2,8(a0)
	addiu	a1,a1,-1
	bne	a1,zero,1b
	addiu	a0,a0,MIDI_SIZE
	addiu	a0,zero,m0
	jr	ra
	sw	a0,G_FUNC(gp)

# in:  a0 midi data (1byte)

m0:
	andi	at,a0,0x80
	beq	at,zero,m1
	nop
	sb	a0,G_STATUS(gp)
	addiu	at,zero,m1
	jr	ra
	sw	at,G_FUNC(gp)
m1:
	lbu	t0,G_STATUS(gp)
	li	t2,MIDI_SIZE
	addiu	t1,t0,-0x80
	andi	t0,t0,0xf	# midi_ch
	sb	t0,G_MIDICH(gp)
	multu	t0,t2
	srl	t1,t1,4
	sll	t1,t1,2
	addiu	t1,t1,m1tbl
	lw	t2,0(t1)
	mflo	t1
	addiu	t1,t1,MIDIWORK
	jr	t2
	sw	t1,G_MIDIWORKPTR(gp)
m1tbl:
	.dc.l	noteoff1,noteon1,keypressure1,control1
	.dc.l	program1,channelpressure1,pitchbend1,other1
noteoff1:
	sll	t1,a0,8
	or	a0,t1,t0	# id
	addiu	t0,zero,noteoff2
	sw	t0,G_FUNC(gp)
	.if	SIM
	j	TestKeyOff
	.else
	j	KeyOff
	.endif
	nop
noteon1:
	sb	a0,MIDI_NOTE(t1)
	addiu	t0,zero,noteon2
	jr	ra
	sw	t0,G_FUNC(gp)
keypressure1:
	addiu	t0,zero,void2
	jr	ra
	sw	t0,G_FUNC(gp)
control1:
	li	at,7
	bne	a0,at,1f
	li	at,10
	j	9f
	addiu	t0,zero,volume2
1:
	bne	a0,at,2f
	li	at,11
	j	9f
	addiu	t0,zero,pan2
2:
	bne	a0,at,3f
	li	at,0x62
	j	9f
	addiu	t0,zero,expression2
3:
	bne	a0,at,4f
	li	at,0x64
	j	9f
	addiu	t0,zero,nrpn2
4:
	bne	a0,at,5f
	li	at,6
	j	9f
	addiu	t0,zero,rpn2
5:
	bne	a0,at,9f
	addiu	t0,zero,void2
	addiu	t0,zero,data2
9:
	jr	ra
	sw	t0,G_FUNC(gp)
program1:
	sb	a0,MIDI_PROG(t1)
	addiu	t0,zero,m0
	jr	ra
	sw	t0,G_FUNC(gp)
channelpressure1:
	addiu	t0,zero,m0
	jr	ra
	sw	t0,G_FUNC(gp)
pitchbend1:
	sb	a0,MIDI_BEND(t1)
	addiu	t0,zero,pitchbend2
	jr	ra
	sw	t0,G_FUNC(gp)
other1:
	beq	t0,zero,1f
	li	at,7
	bne	t0,at,2f
	li	at,15
1:
	andi	t0,a0,0x80
	andi	a0,a0,0x7f
	beql	a0,zero,5f
	addiu	t0,zero,m0
	sw	a0,G_VV(gp)
	bne	t0,zero,4f
	addiu	t0,zero,void3
	jr	ra
	sw	t0,G_FUNC(gp)
4:
	addiu	t0,zero,sysex2
5:
	jr	ra
	sw	t0,G_FUNC(gp)
2:
	bne	t0,at,3f
	addiu	t0,zero,meta2
	sw	zero,G_VV(gp)
	sb	a0,G_META(gp)
	jr	ra
	sw	t0,G_FUNC(gp)
3:
	addiu	t0,zero,m0
	jr	ra
	sw	t0,G_FUNC(gp)
noteoff2:
	addiu	t0,zero,m0
	jr	ra
	sw	t0,G_FUNC(gp)
noteon2:
	lw	t1,G_MIDIWORKPTR(gp)
	lbu	v0,G_MIDICH(gp)
	lbu	a1,MIDI_NOTE(t1)
	addiu	t2,zero,m0
	sll	t0,a1,8
	beq	a0,zero,1f
	sw	t2,G_FUNC(gp)
	li	at,9
	beq	v0,at,2f	# rithem
	sb	a0,MIDI_VELO(t1)
	lbu	a0,MIDI_PROG(t1)
	lbu	v1,MIDI_PAN(t1)
3:
	lbu	a2,MIDI_EXPRESSION(t1)
	lbu	a3,MIDI_VOLUME(t1)
	nop
	multu	a3,a2
	mflo	a3
	multu	a3,a3
	mflo	a3
	srl	a3,a3,20	# velo
	lbu	a2,MIDI_VELO(t1)
	lb	t7,MIDI_BEND(t1)
	lbu	t1,MIDI_BENDSEN(t1)
	nop
	mult	t7,t1
	mflo	t7
	li	at,1
	sb	at,G_STARTED(gp)
	.if	SIM
	j	TestKeyOn
	.else
	j	KeyOn
	.endif
	or	v0,v0,t0
1:
	.if	SIM
	j	TestKeyOff
	.else
	j	KeyOff
	.endif
	or	a0,v0,t0
2:
	lw	t2,G_TONE(gp)
	nop
	lhu	v1,2(t2)
	sll	a1,a1,2
	addu	t2,t2,v1
	addu	a1,t2,a1
	lw	v1,0(a1)
	nop
	beq	v1,zero,4f
	andi	a0,v1,0xff
	srl	a1,v1,8
	andi	a1,a1,0x7f
	srl	v1,v1,16
	j	3b
	andi	v1,v1,0x7f
4:
	jr	ra
	nop
volume2:
	lw	t1,G_MIDIWORKPTR(gp)
	mov	t7,ra
	lbu	t0,MIDI_EXPRESSION(t1)
	j	1f
	sb	a0,MIDI_VOLUME(t1)
expression2:
	lw	t1,G_MIDIWORKPTR(gp)
	mov	t7,ra
	lbu	t0,MIDI_VOLUME(t1)
	sb	a0,MIDI_EXPRESSION(t1)
1:
	multu	a0,t0
	lbu	a0,G_MIDICH(gp)
	mflo	a1
	multu	a1,a1
	mflo	a1
	jal	ChangeVolume
	srl	a1,a1,20
	addiu	t0,zero,m0
	jr	t7
	sw	t0,G_FUNC(gp)
pan2:
	lw	t1,G_MIDIWORKPTR(gp)
	nop
	sb	a0,MIDI_PAN(t1)
	addiu	t0,zero,m0
	jr	ra
	sw	t0,G_FUNC(gp)
rpn2:
	lw	t1,G_MIDIWORKPTR(gp)
	nop
	sb	a0,MIDI_RPN(t1)
	addiu	t0,zero,m0
	jr	ra
	sw	t0,G_FUNC(gp)
nrpn2:
	lw	t1,G_MIDIWORKPTR(gp)
	li	a0,127
	sb	a0,MIDI_RPN(t1)
	addiu	t0,zero,m0
	jr	ra
	sw	t0,G_FUNC(gp)
data2:
	lw	t1,G_MIDIWORKPTR(gp)
	nop
	lbu	t0,MIDI_RPN(t1)
	nop
	bne	t0,zero,1f
	addiu	t0,zero,m0
	andi	a0,a0,0xf
	sb	a0,MIDI_BENDSEN(t1)
1:
	jr	ra
	sw	t0,G_FUNC(gp)
pitchbend2:
	lw	t1,G_MIDIWORKPTR(gp)
	sll	a0,a0,1
	lbu	t0,MIDI_BEND(t1)
	nop
	srl	t0,t0,6
	or	t0,t0,a0
	addiu	t0,t0,-0x80
	sb	t0,MIDI_BEND(t1)
	lbu	a1,MIDI_BENDSEN(t1)
	lbu	a0,G_MIDICH(gp)
	mult	t0,a1
	mflo	a1
	addiu	t0,zero,m0
	j	Bend
	sw	t0,G_FUNC(gp)
sysex2:
	lw	t0,G_VV(gp)
	andi	t2,a0,0x80
	sll	t0,t0,7
	andi	t1,a0,0x7f
	or	t0,t0,t1
	bne	t2,zero,1f
	sw	t0,G_VV(gp)
	addiu	t0,zero,void3
	jr	ra
	sw	t0,G_FUNC(gp)
1:
	jr	ra
	nop
meta2:
	lw	t0,G_VV(gp)
	andi	t1,a0,0x7f
	sll	t0,t0,7
	or	t0,t0,t1
	sw	t0,G_VV(gp)
	andi	a0,a0,0x80
	bne	a0,zero,1f
	nop
	beql	t0,zero,2f
	lbu	t0,G_META(gp)
	lbu	t0,G_META(gp)
	li	at,0x51
	bne	t0,at,3f
	addiu	t0,zero,tempo3
	jr	ra
	sw	t0,G_FUNC(gp)
1:
	jr	ra
	nop
2:
	li	at,0x2f
	beq	t0,at,Next
	addiu	t0,zero,m0
	jr	ra
	sw	t0,G_FUNC(gp)
3:
	addiu	t0,zero,void3
	jr	ra
	sw	t0,G_FUNC(gp)
void2:
	addiu	t0,zero,m0
	jr	ra
	sw	t0,G_FUNC(gp)
void3:
	lw	t0,G_VV(gp)
	nop
	addiu	t0,t0,-1
	beq	t0,zero,1f
	sw	t0,G_VV(gp)
	jr	ra
	nop
1:
	addiu	t0,zero,m0
	jr	ra
	sw	t0,G_FUNC(gp)
tempo3:
	sll	a0,a0,16
	sw	a0,G_TEMPO(gp)
	addiu	t0,zero,tempo4
	jr	ra
	sw	t0,G_FUNC(gp)
tempo4:
	lw	t0,G_TEMPO(gp)
	sll	a0,a0,8
	or	t0,t0,a0
	sw	t0,G_TEMPO(gp)
	addiu	t0,zero,tempo5
	jr	ra
	sw	t0,G_FUNC(gp)
tempo5:
	lw	t0,G_TEMPO(gp)
	nop
	or	t0,t0,a0
	sw	t0,G_TEMPO(gp)
	addiu	t0,zero,m0
	jr	ra
	sw	t0,G_FUNC(gp)

######## CARD ########

	CARD_DATA	=	1
	CARD_CLK	=	2
	CARD_CS	=	4

# u8 CardIO(u8 din);
# in:  a0 din
# out: v0 dout
# use: at,v0,v1,a1

	.macro	cardbit	pos
	srl	v1,a0,\pos
	andi	v1,v1,1
	sb	v1,0(a1)
	ori	at,v1,CARD_CLK
	sb	at,0(a1)
	lbu	at,0(a1)
	sb	v1,0(a1)
	andi	v1,at,1
	sll	v1,v1,\pos
	or	v0,v0,v1
	.endm

CardIO:
	li	a1,IOCARD
	li	v0,0
	cardbit	7
	cardbit	6
	cardbit	5
	cardbit	4
	cardbit	3
	cardbit	2
	cardbit	1
	cardbit	0
	jr	ra
	andi	v0,v0,0xff

# in:  a0 count
# use: at,v0,v1,a0-a3

CardSkip:
	addiu	a2,a0,-1
	mov	a3,ra
1:
	jal	CardIO
	li	a0,0xff
	bne	a2,zero,1b
	addiu	a2,a2,-1
	jr	a3
	nop

# in:  a0 command
#      a1 param
# out: v0 response
# use: at,v0,v1,a0-a3,t0-t2

CardCommand:
	mov	t0,a0
	mov	t1,a1
	mov	t2,ra
1:
	jal	CardIO
	li	a0,0xff
	bne	v0,a0,1b
	nop
	jal	CardIO
	ori	a0,t0,0x40
	jal	CardIO
	srl	a0,t1,24
	jal	CardIO
	srl	a0,t1,16
	jal	CardIO
	srl	a0,t1,8
	jal	CardIO
	mov	a0,t1
	jal	CardIO
	li	a0,0x95
2:
	jal	CardIO
	li	a0,0xff
	beq	v0,a0,2b
	mov	t0,v0
	jal	CardSkip
	li	a0,1
	jr	t2
	mov	v0,t0

# use: at,v0,v1,a0-a3,t0-t3

CardInit:
	mov	t3,ra
	li	a0,IOCARD
	li	at,CARD_CS|CARD_DATA
	sb	at,0(a0)
	li	a2,0x80
1:
	li	at,CARD_CS|CARD_CLK|CARD_DATA
	sb	at,0(a0)
	li	a1,13
2:
	bne	a1,zero,2b
	addiu	a1,a1,-1
	li	at,CARD_CS|CARD_DATA
	sb	at,0(a0)
	li	a1,13
3:
	bne	a1,zero,3b
	addiu	a1,a1,-1
	bne	a2,zero,1b
	addiu	a2,a2,-1
	li	at,CARD_DATA
	sb	at,0(a0)
4:
	jal	CardIO
	li	a0,0xff
	bne	v0,a0,4b
	nop
	li	a0,0
	jal	CardCommand
	li	a1,0
	li	a0,1
5:
	jal	CardCommand
	li	a1,0
	bnel	v0,zero,5b
	li	a0,1
	jr	t3
	sw	zero,G_CARDADR(gp)

# in:  a0 address
# use: at,v0,v1,a0-a3,t0-t4

CardAddress:
	mov	t3,a0
	mov	t4,ra
	lw	a1,G_CARDADR(gp)
	sw	a0,G_CARDADR(gp)
	andi	a1,a1,0x1ff
	beq	a1,zero,1f
	li	a0,0x202
	jal	CardSkip
	subu	a0,a0,a1
1:
	andi	at,t3,0x1ff
	beq	at,zero,2f
	li	a0,17
	li	a1,~0x1ff
	jal	CardCommand
	and	a1,t3,a1
	li	t0,0xfe
3:
	jal	CardIO
	li	a0,0xff
	bne	v0,t0,3b
	nop
	jal	CardSkip
	andi	a0,t3,0x1ff
2:
	jr	t4
	nop

# out: v0 data
# use: at,v0,v1,a0-a3,t0-t4

CardRead:
	lw	t3,G_CARDADR(gp)
	mov	t4,ra
	andi	at,t3,0x1ff
	bne	at,zero,1f
	li	a0,17
	jal	CardCommand
	mov	a1,t3
	li	t0,0xfe
2:
	jal	CardIO
	li	a0,0xff
	bne	v0,t0,2b
	nop
1:
	jal	CardIO
	li	a0,0xff
	mov	t0,v0
	addiu	t3,t3,1
	andi	at,t3,0x1ff
	bne	at,zero,3f
	sw	t3,G_CARDADR(gp)
	jal	CardSkip
	li	a0,2
3:
	jr	t4
	mov	v0,t0

# use: at,v0,v1,a0-a3,t0-t6

CardRead2:
	mov	t5,ra
	jal	CardRead
	nop
	jal	CardRead
	mov	t6,v0
	sll	v0,v0,8
	jr	t5
	or	v0,t6,v0

# use: at,v0,v1,a0-a3,t0-t6

CardRead4:
	mov	t5,ra
	jal	CardRead
	nop
	jal	CardRead
	mov	t6,v0
	sll	v0,v0,8
	jal	CardRead
	or	t6,t6,v0
	sll	v0,v0,16
	jal	CardRead
	or	t6,t6,v0
	sll	v0,v0,24
	jr	t5
	or	v0,t6,v0

# in:  a0 n
# use: at,v0,v1,a0-a3,t0-t6

Discard:
	addiu	t5,a0,-1
	mov	t6,ra
1:
	jal	CardRead
	nop
	bne	t5,zero,1b
	addiu	t5,t5,-1
	jr	t6
	nop

######## FILE ########

# use: at,v0,v1,a0-a3,t0-t8

FileInit:
	mov	t7,ra
	jal	CardInit
	nop
	jal	CardAddress
	li	a0,0x1c6
	jal	CardRead4	# first sector number
	nop
	addiu	t8,v0,1	# w+1
	sll	a0,v0,9
	jal	CardAddress
	ori	a0,a0,0xd
	jal	CardRead	# sector per cluster
	nop
	sll	v0,v0,9
	addiu	v0,v0,-1
	sh	v0,G_CLUSTERMASK(gp)
	jal	Discard
	li	a0,8
	sll	t8,t8,9
	sw	t8,G_FAT(gp)
	jal	CardRead2
	nop
	sll	v0,v0,10
	addu	t8,t8,v0
	sw	t8,G_RDE(gp)
	sw	zero,G_FILELEN(gp)
	jr	t7
	sw	zero,G_FILEOFS(gp)

# in:  a0 c
# use: at,v0,v1,a0-a3,t0-t5

	LAST_CLUSTER	=	0xfff7
	RDE_LEN	=	0x4000

SetCluster:
	li	a1,LAST_CLUSTER
	subu	at,a0,a1
	bgez	at,1f
	mov	t5,ra
	lw	a1,G_RDE(gp)	# adr
	beq	a0,zero,2f
	addiu	a0,a0,-2
	lhu	a2,G_CLUSTERMASK(gp)
	addiu	a1,a1,RDE_LEN
	addiu	a2,a2,1
	multu	a2,a0
	mflo	a0
	addu	a1,a1,a0
2:
	jal	CardAddress
	mov	a0,a1
	jr	t5
	nop
1:
	jr	ra
	nop

# in:  a0 len
#      s4: sCount (first call only)
#      s5: sCluster (first call only)
# use: at,v0,v1,a0-a3,t0-t8,s0-s5

Seek:
	mov	t8,ra
	mov	s0,a0	# len
	li	s1,0	# ofs
	mov	s2,s5	# mycluster
	jal	SetCluster
	mov	a0,s2
	j	1f
	nop
2:
	jal	CardRead
	nop
	beq	v0,zero,3f
	li	at,5
	beq	v0,at,3f
	li	at,0x2e
	beq	v0,at,3f
	li	at,0xe5
	beq	v0,at,3f
	li	at,0x5f
	bne	v0,at,4f
	nop
3:
	jal	Discard
	li	a0,31
	j	5f
	nop
4:
	jal	Discard
	li	a0,7
	jal	CardRead4
	nop
	li	t0,0x10ffffff
	and	t7,v0,t0
	jal	Discard
	li	a0,14
	jal	CardRead2
	nop
	beq	v0,zero,6f
	mov	s5,v0
	li	at,0x44494d
	bnel	t7,at,8f
	srl	t7,t7,28
	addiu	s4,s4,-1
	beq	s4,zero,9f
6:
	srl	t7,t7,28
8:
	andi	at,t7,1
	beq	at,zero,7f
	nop
	lw	s3,G_CARDADR(gp)
	addiu	sp,sp,-20
	sw	s0,0(sp)
	sw	s1,4(sp)
	sw	s2,8(sp)
	sw	s3,12(sp)
	sw	t8,16(sp)
	jal	Seek
	li	a0,0
	lw	s0,0(sp)
	lw	s1,4(sp)
	lw	s2,8(sp)
	lw	s3,12(sp)
	lw	t8,16(sp)
	beq	s4,zero,9f
	addiu	sp,sp,20
	jal	CardAddress
	mov	a0,s3
7:
	jal	Discard
	li	a0,4
5:
	addiu	s1,s1,0x20
	bne	s0,zero,1f
	nop
	lhu	t0,G_CLUSTERMASK(gp)
	nop
	and	t0,s1,t0
	bne	t0,zero,1f
	nop
	lw	a0,G_FAT(gp)
	sll	a1,s2,1
	jal	CardAddress
	addu	a0,a0,a1
	jal	CardRead2
	nop
	mov	s2,v0
	li	at,LAST_CLUSTER
	subu	v0,v0,at
	bgez	v0,9f
	nop
	jal	SetCluster
	mov	a0,s2
1:
	beq	s0,zero,2b
	subu	at,s1,s0
	bltz	at,2b
	nop
9:
	jr	t8
	nop

# in:  a0 file number
# use: at,v0,v1,a0-a3,t0-t9,s0-s5

FileOpen:
	mov	t9,ra
	addiu	s4,a0,1
	jal	printhex
	nop
	li	s5,0
	jal	Seek
	li	a0,RDE_LEN
	bne	s4,zero,1f
	sh	s5,G_CLUSTER(gp)
	jal	CardRead4
	sw	zero,G_FILEOFS(gp)
	sw	v0,G_FILELEN(gp)
	jal	SetCluster
	mov	a0,s5
1:
	jr	t9
	nop

# out: v0 data
# use: at,v0,v1,a0-a3,t0-t8

	.if	CARD
GetChar:
	lw	t8,G_FILEOFS(gp)
	lw	a1,G_FILELEN(gp)
	mov	t7,ra
	subu	at,t8,a1
	bgez	at,2f
	nop
	beql	t8,zero,1f
	addiu	t8,t8,1
	lhu	a1,G_CLUSTERMASK(gp)
	nop
	and	at,t8,a1
	bnel	at,zero,1f
	addiu	t8,t8,1
	lhu	a0,G_CLUSTER(gp)
	lw	a1,G_FAT(gp)
	sll	a0,a0,1
	jal	CardAddress
	addu	a0,a0,a1
	jal	CardRead2
	nop
	sh	v0,G_CLUSTER(gp)
	li	t0,LAST_CLUSTER
	subu	at,v0,t0
	bgez	at,2f
	nop
	jal	SetCluster
	mov	a0,v0
	addiu	t8,t8,1
1:
	sw	t8,G_FILEOFS(gp)
	j	CardRead
	mov	ra,t7
2:
	jr	t7
	li	v0,-1

# out: v0 buffer full or end of file
# use: at,v0,v1,a0-a3,t0-t9,s0

idle:
	lw	s0,G_BUFN(gp)
	li	at,BUFN
	beq	s0,at,1f
	mov	t9,ra
	jal	GetChar
	nop
	bltz	v0,1f
	nop
	lw	a0,G_BUFWP(gp)
	li	at,BUF+BUFN
	bne	a0,at,2f
	addiu	s0,s0,1
	li	a0,BUF
2:
	sb	v0,0(a0)
	addiu	a0,a0,1
	sw	a0,G_BUFWP(gp)
	sw	s0,G_BUFN(gp)
	jr	t9
	li	v0,0
1:
	jr	t9
	li	v0,1
	.endif

# use: at,v0,v1,a0-a3,t0-t9,s0

FillBuffer:
	mov	s1,ra
	.if	CARD
	sw	zero,G_BUFN(gp)
	li	a0,BUF
	sw	a0,G_BUFWP(gp)
	sw	a0,G_BUFRP(gp)
1:
	jal	idle
	nop
	beq	v0,zero,1b
	nop
	.endif
	jr	s1
	nop

######## SMF ########

# u8 getc();
# out: v0 character
# use: at,a0,a1

getc:
	.if	CARD
	lw	a0,G_BUFN(gp)
	lw	a1,G_BUFRP(gp)
	beq	a0,zero,Next
	nop
	li	at,BUF+BUFN
	bne	a1,at,1f
	addiu	a0,a0,-1
	li	a1,BUF
1:
	lbu	v0,0(a1)
	addiu	a1,a1,1
	sw	a0,G_BUFN(gp)
	jr	ra
	sw	a1,G_BUFRP(gp)
	.else
	lw	a0,G_MIDIPTR(gp)
	lw	a1,G_MIDILIM(gp)
	lbu	v0,0(a0)
	subu	at,a0,a1
	bgez	at,End
	addiu	a0,a0,1
	jr	ra
	sw	a0,G_MIDIPTR(gp)
	.endif

SMFHeader:
	.if !SIM
	.if	CARD
	mov	s1,ra
	li	s0,12-1
1:
	jal	getc
	nop
	bne	s0,zero,1b
	addiu	s0,s0,-1
	jal	getc
	nop
	sll	s0,v0,8
	jal	getc
	nop
	or	s0,s0,v0
	sh	s0,G_TIMEBASE(gp)
	li	s0,8-1
2:
	jal	getc
	nop
	bne	s0,zero,2b
	addiu	s0,s0,-1
	jr	s1
	nop
	.else
	la	a0,midi+22
	sw	a0,G_MIDIPTR(gp)
	ori	t0,zero,midiend
	sw	t0,G_MIDILIM(gp)
	lbu	t0,-10(a0)
	lbu	t1,-9(a0)
	sll	t0,t0,8
	or	t0,t0,t1
	jr	ra
	sh	t0,G_TIMEBASE(gp)
	.endif
	.endif

SMFPlay:
	li	s0,0
1:
	jal	getc
	nop
	sll	s0,s0,7
	andi	s1,v0,0x7f
	or	s0,s0,s1
	andi	s1,v0,0x80
	bne	s1,zero,1b
	nop
	beq	s0,zero,2f
	nop
#
	.if	!SIM
	li	a0,0x100	# avoid click noise
	jal	TimerSet
	li	a1,0
	jal	FMStart
	nop
7:
	jal	TimerCheck
	nop
	beq	v0,zero,7b
	nop
	jal	FMStop
	nop
	lw	s1,G_ACTIVECH(gp)
	nop
	j	4f
	nop
6:
	nop
	beql	s2,zero,5f
	lw	s1,0(s1)
	sb	zero,CH_FLAG(s1)
	mov	a0,s1
	lbu	a1,CH_PROG(s1)
	lbu	a2,CH_NOTE(s1)
	lbu	a3,CH_VELO(s1)
	jal	ChannelOn
	nop
	lw	s1,0(s1)
5:
	nop
4:
	bnel	s1,zero,6b
	lbu	s2,CH_FLAG(s1)
#
	lw	t1,G_TEMPO(gp)
	lhu	a1,G_TIMEBASE(gp)
	srl	t1,t1,4	# avoid overflow
	multu	t1,s0
	.if NATIVE_DIV
	mflo	a0
	.dc.l	4<<21|5<<16|0x1b	# divu
	jal	TimerGetAdjust
	mflo	a0
	.else
	jal	Div
	mflo	a0
	jal	TimerGetAdjust
	mov	a0,v0
	.endif
	lbu	at,G_STARTED(gp)
	nop
	bnel	at,zero,8f
	sll	a0,a0,4	# after started, restore original value
8:
	jal	TimerSet
	mov	a1,v0
	jal	FMStart
	nop
3:
	.if CARD
	jal	idle
	nop
	.endif
	jal	RestartCheck
	nop
	jal	TimerCheck
	nop
	beq	v0,zero,3b
	li	a0,0
	jal	TimerSet
	li	a1,0
	jal	FMStop
	nop
	.endif
2:
	jal	getc
	nop
	lw	t0,G_FUNC(gp)
	mov	a0,v0
	jalr	t0
	nop
	lw	t0,G_FUNC(gp)
	addiu	at,zero,m0
	bne	t0,at,2b
	nop
	j	SMFPlay
	nop

RestartCheck:
	mov	a3,ra
	jal	ButtonPoll
	nop
	lbu	v0,G_KEY(gp)
	nop
	andi	v0,v0,PB_ALL
	bne	v0,zero,End
	nop
	jr	a3
	nop

	RESTART_WAIT	=	50

Next:
	li	v0,PB_NEXT
	sb	v0,G_KEY(gp)
	sb	v0,G_KEYON(gp)
	# fall to End
End:
	jal	FMStop
	nop
	jal	NoteOffAll
	nop
	jal	FMStart
	nop
5:
	lbu	v0,G_KEYON(gp)
	lhu	v1,G_FILENUM(gp)
	andi	at,v0,PB_RESET
	beq	at,zero,4f
	andi	at,v0,PB_PREV
	li	a0,IOBASE
	lhu	v1,0(a0)
	nop
	andi	v1,v1,0xfff
	j	2f
	xori	v1,v1,0xf00	# upper
4:
	beq	at,zero,1f
	andi	at,v0,PB_NEXT
	beq	v1,zero,6f
	nop
	j	2f
	addiu	v1,v1,-1
1:
	beq	at,zero,3f
	addiu	v1,v1,1
2:
	sh	v1,G_FILENUM(gp)
3:
	lbu	v0,G_KEY(gp)
	nop
	andi	v0,v0,PB_ALL
	bnel	v0,zero,6f
	li	t7,RESTART_WAIT
6:
	li	a0,10000	# 10mS
	jal	TimerSet
	li	a1,0
7:
	jal	TimerCheck
	nop
	beq	v0,zero,7b
	nop
	jal	ButtonPoll
	nop
	bne	t7,zero,5b
	addiu	t7,t7,-1
	j	restart
	nop

######## UTIL ########

# use: v0,v1,a0

ButtonPoll:
	li	a0,IO_PBUTTON
	lhu	v0,0(a0)
	lbu	v1,G_KEY(gp)
	srl	v0,v0,3
	sb	v0,G_KEY(gp)
	nor	v1,v1,v1
	and	v0,v0,v1
	jr	ra
	sb	v0,G_KEYON(gp)

# in:  a0 wait time (uS)
#      a1 adjust
# use: v0,a0

TimerSet:
	li	v0,CLOCK_FREQ
	multu	a0,v0
	li	v0,TIMER
	lhu	v0,0(v0)
	mflo	a0
	sh	v0,G_CURTIME(gp)
	addu	a0,a0,v0
	subu	a0,a0,a1
	bltzl	a0,1f
	mov	a0,v0
1:
	jr	ra
	sw	a0,G_TARGETTIME(gp)

# use: at,t0-t3
# out: v0 0...no 1...yes

TimerCheck:
	li	t0,TIMER
	lhu	t0,0(t0)
	lhu	t1,G_CURTIME(gp)
	sh	t0,G_CURTIME(gp)
	lw	t2,G_TARGETTIME(gp)
	subu	at,t0,t1
	bgez	at,1f
	lui	t3,1
	subu	t2,t2,t3
	sw	t2,G_TARGETTIME(gp)
	bltz	t2,2f
1:
	srl	t3,t2,16
	bne	t3,zero,3f
	subu	at,t0,t2
	bgez	at,2f
	nop
3:
	jr	ra
	li	v0,0
2:
	jr	ra
	li	v0,1

# out: v0 adjust value
# use: v0,t0,t1

TimerGetAdjust:
	lhu	t0,G_CURTIME(gp)
	lw	t1,G_TARGETTIME(gp)
	nop
	subu	v0,t0,t1
	jr	ra
	andi	v0,v0,0xffff

# in:  a0 digit (0-f)
# out: v0 7 seg data
# use: v0,a0

dec7seg:
	la	v0,data7seg
	andi	a0,a0,0xf
	addu	a0,a0,v0
	lbu	v0,0(a0)
	jr	ra
	nop
data7seg:
	.dc.b	0x3f,0x06,0x5b,0x4f,0x66,0x6d,0x7d,0x07
	.dc.b	0x7f,0x6f,0x77,0x7c,0x39,0x5e,0x79,0x71

# in:  a0  hex data (0-ffff)
# use: v0,a0-a3

printhex:
	mov	a2,ra
	jal	dec7seg
	mov	a1,a0
	mov	a3,v0
	jal	dec7seg
	srl	a0,a1,4
	sll	v0,v0,8
	or	a3,a3,v0
	jal	dec7seg
	srl	a0,a1,8
	sll	v0,v0,16
	or	a3,a3,v0
	jal	dec7seg
	srl	a0,a1,12
	sll	v0,v0,24
	or	a3,a3,v0
	jr	a2
	sw	a3,G_7SEG(gp)

# in: a0  a
#     a1  b
# out: v0 a/b
# use: at,v0,v1,a0,a1

Div:
	beq	a1,zero,1f
	li	v0,0	# v
	li	v1,-1	# i
	j	3f
	subu	a1,zero,a1
2:
	sll	a1,a1,1
3:
	addu	at,a0,a1
	bgez	at,2b
	addiu	v1,v1,1
	j	4f
	nop
5:
	addu	at,a0,a1
	bltz	at,6f
	sra	a1,a1,1
	ori	v0,v0,1
	addu	a0,zero,at
6:
	sll	v0,v0,1
4:
	bne	v1,zero,5b
	addiu	v1,v1,-1
1:
	jr	ra
	nop

	.if	SIM
TestKeyOn:
	li	t0,IOBASE
	sb	zero,0x7fff(t0)
	jr	ra
	nop
TestKeyOff:
	li	t0,IOBASE
	sb	zero,0x7ffe(t0)
	jr	ra
	nop
	.endif

### data

	.include "deltatable"
	.if	!CARD && !SIM
	.include "tone.inc"
	.include "midi.inc"
	.endif
	.if	BOOTABLE || SIM
	.org	TONEADR
	.include "tone.inc"
	.endif
